/*******************************************************************************
* Implementation of identifier information and scope administration actions,
* used in a SLADE parser specification
*
* JHG sep97 Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#include "idscope.h"

static DefRec	i_undeclared= {
					readonly_access, dim0, const_kind, unknown_type, 0,
					0, no_dva, 0, false, 0, 0, &i_undeclared, &i_undeclared
				};
def_ptr			undeclared= &i_undeclared;	/* empty definition record */
static DefRec	i_null_def= {
					readonly_access, dim0, ptr_kind, unknown_type, "NULL",
					0, no_dva, 0, false, 0, 0, &i_undeclared, &i_undeclared
				};
def_ptr			null_def= &i_null_def;		/* null pointer definition record */

typedef struct idrec {
	def_ptr				def;
	boolean				set_or_use;
	struct idrec		*next;
} IDREC;

typedef struct idnode {
	string				name;
	IDREC				*decls;		/* list of identifiers sharing one name */
	struct idnode		*left, *right;
} IDNODE;

typedef struct scoperec {
	integer				depth;
	IDREC				*idlist;	/* list of identifiers within one scope */
	struct scoperec		*next;
} SCOPEREC;

#define ACCESS_TEXTLEN	30
#define DEF_PTR_TEXTLEN	80

static SCOPEREC			*ScopeList= 0;
static IDNODE			*IdTree= 0;

#ifdef Debugger
void
DeleteTree(IDNODE *node)
{
	if (node == 0)
		return;

	DeleteTree(node->left);
	DeleteTree(node->right);
	free((char *)node);
}

void
i_reset_idscope(void)
{
	while (ScopeList)
		exit_scope();

	DeleteTree(IdTree);
	IdTree= 0;
}
#endif /* Debugger */

char *
i_e_access_text(e_access a)
{
	static char		buf[ACCESS_TEXTLEN];
	e_access		a2;

	if ((a & readonly_access) == readonly_access) {
		strcpy(buf, "readonly, ");
		a2= a & ~readonly_access;
	} else {
		buf[0]= 0;
		a2= a;
	}

	switch (a2) {
	case address_access:	strcat(buf, "address"); break;
	case direct_access:		strcat(buf, "direct"); break;
	case indirect_access:	strcat(buf, "indirect"); break;
	default:				strcat(buf, "unknown");
	}
	return buf;
}

char *
i_def_ptr_text(def_ptr def)
{
	static char buf[DEF_PTR_TEXTLEN];
	int			len, i;

	if (def == 0)
		return "NULL";

	if (def == undeclared)
		return "undeclared";

	if (def == null_def)
		return "null_def";

	if (def->name) {
		sprintf(buf, "'%.*s' ", 10, def->name);
		if (def->dva != no_dva)
			sprintf(buf+strlen(buf), "%d, %d, ", def->sn, def->dva);
		len= strlen(buf);
	} else
		len= 0;

	sprintf(buf+len, "%s, ", i_e_type_text(def->type));

	if ((def->access&direct_access) != direct_access)
		sprintf(buf+strlen(buf), "%s, ", i_e_access_text(def->access));

	strcat(buf, i_e_kind_text(def->kind));

	for (i= 0; i < def->dim; i++)
		if (def->bound)
			sprintf(buf+strlen(buf), "[%d..%d]", def->bound[i].low
			, def->bound[i].low + def->bound[i].size - 1);
		else
			strcat(buf, "[]");

	if (def->set_or_use && *(def->set_or_use))
		strcat(buf, ", set");

	return buf;
}

char *
i_e_kind_text(e_kind k)
{
	switch (k) {
	case avar_kind:		return "avar";
	case const_kind:	return "const";
	case proc_kind:		return "proc";
	case ptr_kind:		return "ptr";
	case rec_kind:		return "rec";
	case rpar_kind:		return "rpar";
	case svar_kind:		return "svar";
	case tdecl_kind:	return "tdecl";
	case vpar_kind:		return "vpar";
	default:			;
	}
	return "unknown";
}

char *
i_e_type_text(e_type t)
{
	switch (t) {
	case bool_type:		return "bool";
	case char_type:		return "char";
	case int_type:		return "int";
	case no_type:		return "no type";
	case string_type:	return "string";
	default:			;
	}
	return "unknown";
}

void
assign_access(e_access *dest, e_access src)
{
	*dest= src;
}

void
assign_definition(def_ptr *dest, def_ptr src)
{
	*dest= src;
}

void
assign_kind(e_kind *dest, e_kind src)
{
	*dest= src;
}

void
assign_type(e_type *dest, e_type src)
{
	*dest= src;
}

void
clear_access(e_access *dest)
{
	*dest= unknown_access;
}

void
clear_definition(def_ptr *dest)
{
	*dest= undeclared;
}

void
clear_kind(e_kind *dest)
{
	*dest= unknown_kind;
}

void
clear_type(e_type *dest)
{
	*dest= unknown_type;
}

void
check_access(e_access a1, e_access a2, boolean equal)
{
	if (a1 != a2) {
		if (equal) {
			char	buf[ACCESS_TEXTLEN];

			strcpy(buf, i_e_access_text(a1));
			i_report(i_EACCONF, buf, i_e_access_text(a2));
		}
	} else if (!equal)
		i_report(i_EACCNOTA, i_e_access_text(a1));
}

void
check_dim(integer dim1, integer dim2, boolean equal)
{
	if (dim1 != dim2) {
		if (equal)
			i_report(i_EDIMCONF, dim1, dim2);
	} else if (!equal)
		i_report(i_EDIMNOTA, dim1);
}

void
check_kind(e_kind k1, e_kind k2, boolean equal)
{
	if (k1 == unknown_kind || k2 == unknown_kind)
		return; /*to prevent error messages 16/6/98*/

	if (k1 != k2) {
		if (equal)
			i_report(i_EKINDCONF, i_e_kind_text(k1), i_e_kind_text(k2));
	} else if (!equal)
		i_report(i_EKINDNOTA, i_e_kind_text(k1));
}

void
check_type(e_type t1, e_type t2, boolean equal)
{
	if (t1 == unknown_type || t2 == unknown_type)
		return; /*to prevent error messages in null pointer assignment 16/6/98*/

	if (t1 != t2) {
		if (equal)
			i_report(i_ETYPECONF, i_e_type_text(t1), i_e_type_text(t2));
	} else if (!equal)
		i_report(i_ETYPENOTA, i_e_type_text(t1));
}

static void
boundary_compare(int size1, int size2, boolean equal)
{
	if (size1 != size2) {
		if (equal) {
			char	asize1[10], asize2[10];

			sprintf(asize1, "%d", size1);
			sprintf(asize2, "%d", size2);
			i_report(i_EBNDCONF, asize1, asize2);
		}
	} else if (!equal)
		i_report(i_EBNDNOTA, size1);
}

static void
child_boundary_check(def_ptr def, def_ptr child, boolean equal)
{
	def_ptr	cdef;
	int		i, nchildren;

	for (i= 0; i < def->dim; i++) {
		if (child == undeclared) {
			check_dim(i, def->dim, equal);
			return;
		}

		nchildren= 0;
		for (cdef= child; cdef != undeclared; cdef= cdef->sibling)
			nchildren++;

		if (def->bound == 0) {
			char	asize[10];

			sprintf(asize, "%d", nchildren);
			i_report(i_EBNDCONF, asize, "??");
			return;
		}

		boundary_compare(nchildren, def->bound[i].size, equal);
		child= child->child;
	}
}

static void
boundary_check(def_ptr def1, def_ptr def2, boolean equal)
{
	int		i;
	char	asize[10];

	if (def1->dim == 0 && def2->dim == 0)
		return;

	if (def1->bound == 0 && def2->bound == 0) {
		i_report(i_EBNDCONF, "??", "??");
		return;
	}
	if (def1->bound == 0) {
		sprintf(asize, "%d", def2->bound[0].size);
		i_report(i_EBNDCONF, "??", asize);
		return;
	}
	if (def2->bound == 0) {
		sprintf(asize, "%d", def1->bound[0].size);
		i_report(i_EBNDCONF, asize, "??");
		return;
	}

	boundary_compare(def1->bound[0].size, def2->bound[0].size, equal);

	if (def1->child != undeclared) {
		if (def2->child != undeclared) {
			def_ptr	cdef1, cdef2;

			cdef1= def1->child;
			cdef2= def2->child;
			do {
				boundary_check(cdef1, cdef2, equal);
				cdef1= cdef1->sibling;
				cdef2= cdef2->sibling;
			} while (cdef1 != undeclared && cdef2 != undeclared);
		} else
			child_boundary_check(def2, def1->child, equal);
	} else if (def2->child == undeclared) {

		check_dim(def1->dim, def2->dim, equal);

		for (i= 1; i < def1->dim && i < def2->dim; i++)
			boundary_compare(def1->bound[i].size, def2->bound[i].size, equal);
	} else
		child_boundary_check(def1, def2->child, equal);
}

static e_kind
derive_kind(def_ptr def)
{
	e_kind	k;

	k= def->kind;
	switch (k) {
	case const_kind:
	case rpar_kind:
	case svar_kind:
	case vpar_kind:
		k= svar_kind;
		break;

	case rec_kind:
	case avar_kind:
		k= (def->dim == 0) ? svar_kind : avar_kind;
		break;

	case ptr_kind:
		if ((def->access&indirect_access) == indirect_access)
			k= svar_kind;
		break;

	case proc_kind:
		if (def->type != no_type)
			k= svar_kind;
		break;

	default:
		;
	}

	if (k == svar_kind && (def->access&address_access) == address_access)
		k= ptr_kind;

	return k;
}

void
check_definition(def_ptr def1, def_ptr def2, boolean equal)
{
	e_kind	k1, k2;

	if (def1 == undeclared || def2 == undeclared)
		return;

	if (def1->dim > 0 && def2->dim > 0) {
		def_ptr cdef, cdef1, cdef2;
		if (def1->kind == avar_kind) {
			if (def2->kind == rec_kind
			|| (def2->access&readonly_access) == readonly_access)
				for (cdef= def2->child; cdef != undeclared; cdef= cdef->sibling)
					check_type(def1->type, cdef->type, equal);
			else if (def2->kind == avar_kind)
				check_type(def1->type, def2->type, equal);
		} else if (def1->kind == rec_kind) {
			if (def2->kind == rec_kind
			|| (def2->access&readonly_access) == readonly_access) {
				cdef1= def1->child;
				cdef2= def2->child;
				while (cdef1 != undeclared && cdef2 != undeclared) {
					check_type(cdef1->type, cdef2->type, equal);
					cdef1= cdef1->sibling;
					cdef2= cdef2->sibling;
				}
			} else if (def2->kind == avar_kind)
				for (cdef= def1->child; cdef != undeclared; cdef= cdef->sibling)
					check_type(cdef->type, def2->type, equal);
		}
	} else
		check_type(def1->type, def2->type, equal);

	boundary_check(def1, def2, equal);

	k1= derive_kind(def1);
	k2= derive_kind(def2);
	check_kind(k1, k2, equal);
}

void
chk_def_access(e_access a, def_ptr def, boolean equal)
{
	check_access(a, def->access, equal);
}

void
chk_def_dim(integer dim, def_ptr def, boolean equal)
{
	check_dim(dim, def->dim, equal);
}

void
chk_def_kind(e_kind k, def_ptr def, boolean equal)
{
	check_kind(k, def->kind, equal);
}

void
chk_def_type(e_type t, def_ptr def, boolean equal)
{
	check_type(t, def->type, equal);
}

void
pointer_check(def_ptr defp, def_ptr def)
{
	if (defp->kind == ptr_kind
	&& (def->access&address_access) == address_access
	&& defp->sn < def->sn)
		i_report(i_WDANGLE);
}

void
positive_check(integer n, e_kind k)
{
	if (n > 0)
		i_report(i_ETOOFEW, (k == proc_kind) ? "arguments" : "indices");
}

void
zero_check(integer n, e_kind k)
{
	if (n == 0)
		i_report(i_ETOOMANY, (k == proc_kind) ? "arguments" : "indices");
}

void
first_parm(def_ptr *def_par, integer *npars, def_ptr def_proc)
{
	*def_par= def_proc->child;
	*npars= def_proc->nchildren;
}

void
next_parm(def_ptr *def)
{
	*def= (*def)->sibling;
}

void
arg_check(def_ptr argdef, def_ptr pardef)
{
	chk_def_type(argdef->type, pardef, true);
	if ((argdef->access&readonly_access) == readonly_access
	&& pardef->kind == rpar_kind)
		i_report(i_EILLREF, argdef->name ? argdef->name : "expression");
}

void
arg_access(e_access *a, def_ptr pardef)
{
	if (pardef->kind == rpar_kind)
		*a= address_access;
	else
		*a= unknown_access;
}

static def_ptr
new_definition(void)
{
	def_ptr	def;

	if ((def= (DefRec *)malloc(sizeof(DefRec))) == 0) {
		i_report(i_ENOMEM, __LINE__, __FILE__);
		return undeclared;
	}
	def->name= 0;
	def->bound= 0;
	def->delete= true;
	def->set_or_use= 0;
	def->nchildren= 0;
	def->child= undeclared;
	def->sibling= undeclared;
	return def;
}

static void
print_definition(def_ptr def)
{
	if (def == undeclared)
		return;

	puts(i_def_ptr_text(def));
	print_definition(def->child);
	print_definition(def->sibling);
}

void
create_definition(def_ptr *def, integer v, e_type t, e_kind k, integer n)
{
	if ((*def= new_definition()) == undeclared)
		return;

	if (ScopeList)
		(*def)->sn= ScopeList->depth;
	else
		(*def)->sn= 0;

	(*def)->dva=	v;
	(*def)->dim=	n;
	(*def)->kind=	k;
	(*def)->type=	t;

	switch (k) {
	case const_kind:
	case proc_kind:
	case tdecl_kind:
		(*def)->access= readonly_access;
		break;

	case rpar_kind:
		(*def)->access= indirect_access;
		break;

	default:
		(*def)->access= direct_access;
	}
#ifdef DEBUG
	printf("create: ");
	print_definition(*def);
#endif /* DEBUG */
}

void
add_def_bound(def_ptr def, int low, int up)
{
	if (low == up)
		return;

	if (low > up) {
		i_report(i_ERANGE, low, up);
		return;
	}

	if (def->bound == 0)
		def->bound= (BoundRec *)malloc(sizeof(BoundRec));
	else
		def->bound= (BoundRec *)realloc(def->bound
		, (def->dim + 1)*sizeof(BoundRec));

	if (def->bound == 0) {
		i_report(i_ENOMEM, __LINE__, __FILE__);
		return;
	}

	(def->bound[def->dim]).low= low;
	(def->bound[def->dim]).size= up - low + 1;
	def->dim++;
}

static void
delete_siblings(def_ptr def)
{
	if (def == undeclared)
		return;

	delete_siblings(def->sibling);
	def->sibling->delete= true;
	delete_definition(def->sibling);
	def->sibling= undeclared;
}

static void
delete_children(def_ptr def)
{
	if (def->nchildren > 0) {
		delete_siblings(def->child);
		def->child->delete= true;
		delete_definition(def->child);
		def->nchildren= 0;
		def->child= undeclared;
	}
}

void
delete_definition(def_ptr def)
{
	if (def == undeclared || def == null_def)
		return;

	if (!def->delete)
		return;

#ifdef DEBUG
	printf("delete: ");
	print_definition(def);
#endif /* DEBUG */
	delete_children(def);

	if (def->bound) {
		free(def->bound);
		def->bound= 0;
	}

	if (def->name) {
		free(def->name);
		def->name= 0;
	}
	free(def);
}

void
result_access(e_access *ra, e_access a, def_ptr def)
{
	if (a != def->access) {
		chk_def_access(a, def, true);
		*ra= unknown_access;
	} else
		*ra= a;
}

void
result_definition(def_ptr *rdef, def_ptr def1, def_ptr def2)
{
	check_definition(def1, def2, true);
	delete_definition(def2);
	if (def1->name) {
		free(def1->name);
		def1->name= 0;
	}
	*rdef= def1;
}

void
result_kind(e_kind *rk, e_kind k, def_ptr def)
{
	if (k != def->kind) {
		chk_def_kind(k, def, true);
		*rk= unknown_kind;
	} else
		*rk= k;
}

void
result_type(e_type *rt, e_type t, def_ptr def)
{
	if (t != def->type) {
		chk_def_type(t, def, true);
		*rt= unknown_type;
	} else
		*rt= t;
}

void
if_definition(def_ptr *result, def_ptr then_part, def_ptr else_part)
{
	if (then_part->type != else_part->type)
		put_def_type(no_type, then_part);

	if (then_part->dim != else_part->dim)
		put_def_dim(dim0, then_part);

	delete_definition(else_part);
	*result= then_part;
}

/***************************************************/
/* Procedures for the identifier tree              */
/*                                                 */
/* The identifier table is represented             */
/* by a binary tree. The identifiers are in        */
/* alphabetical order by traversing the tree       */
/* in order.                                       */
/*                                                 */
/***************************************************/

void
enter_scope(void)
{
	SCOPEREC *scope;

	if ((scope= (SCOPEREC *)malloc(sizeof(SCOPEREC))) == 0) {
		i_report(i_ENOMEM, __LINE__, __FILE__);
		return;
	}
	if (ScopeList == 0)
		scope->depth= 1;
	else
		scope->depth= ScopeList->depth + 1;
	scope->idlist= 0;
	scope->next= ScopeList;
	ScopeList= scope;
#ifdef DEBUG
	printf("enter_scope(%d)\n", ScopeList->depth);
#endif /* DEBUG */
}

void
current_scope(integer *level)
{
  	if (ScopeList)
		*level= ScopeList->depth;
	else
		*level= 0;
}

static IDNODE *
search(IDNODE *root, string name)
{
	int		match;

	if (root == 0 || root->name == 0 || name == 0)
		return 0;

	match= strcmp(name, root->name);
	if (match < 0)		/* less */
		return search(root->left, name);

	if (match > 0)		/* greater */
		return search(root->right, name);

	return root;		/* equal */
}

void
exit_scope(void)
{
	SCOPEREC	*scope;
	IDREC		*idptr, *next= 0, *iddecl;
	IDNODE		*idnode;
	def_ptr		def;

	if (ScopeList == 0) {
		i_report(i_ENOSCOPE);
		return;
	}
#ifdef DEBUG
	printf("exit_scope(%d)\n", ScopeList->depth);
#endif /* DEBUG */

	for (idptr= ScopeList->idlist; idptr; idptr= next) {
		def= idptr->def;

		/* search identifiers node in symbol tree */
		idnode= search(IdTree, def->name);

		/* if found remove first declaration from node's declaration list */
		if (idnode || idnode->decls) {
			iddecl= idnode->decls;
			if (def->delete && !iddecl->set_or_use) {
				if (def->kind != proc_kind)
					i_report(i_WNOTUSED, def->name);
				else
					i_report(i_WNOCALL, def->name);
			}

			idnode->decls= iddecl->next;
			free(iddecl);
		}

		delete_definition(def);

		next= idptr->next;
		free(idptr);
	}

	scope= ScopeList;
	ScopeList= scope->next;
	free(scope);
}

static IDNODE *
insert(IDNODE **root, string name)
{
	int		match;

	if (*root == 0) {
		if ((*root= (IDNODE *)malloc(sizeof(IDNODE))) != 0) {
			(*root)->decls= 0;
			(*root)->left= 0;
			(*root)->right= 0;
			strcpy((*root)->name, name);
		}
		return *root;
	}

	match= strcmp(name, (*root)->name);
	if (match < 0)					/* less */
		return insert(&(*root)->left, name);

	if (match > 0)					/* greater */
		return insert(&(*root)->right, name);

	return *root;					/* equal */
}

void
insert_definition(string name, def_ptr def)
{
	IDNODE	*idnode;
	IDREC	*idscope, *iddecl;

	if (name == 0)
		return;

	if (ScopeList == 0) {
		i_report(i_ENOIDINS, name);
		return;
	}

	idscope= (IDREC *)malloc(sizeof(IDREC));
	iddecl= (IDREC *)malloc(sizeof(IDREC));
	idnode= insert(&IdTree, name);
	if (idnode == 0 || idscope == 0 || iddecl == 0) {
		i_report(i_ENOMEM, __LINE__, __FILE__);
		if (idscope)
			free(idscope);
		if (iddecl)
			free(iddecl);
		return;
	}

	if (def->sn > 0 && idnode->decls && def->sn <= idnode->decls->def->sn) {
		i_report(i_EDBLDEF, name);
		free(idscope);
		free(iddecl);
		return;
	}
	if (def->name)
		free(def->name);
	def->name= strdup(idnode->name);
	def->set_or_use= &(iddecl->set_or_use);

	iddecl->def= def;
	iddecl->set_or_use= false;
	iddecl->next= idnode->decls;		/* push decl list of found idnode */
	idnode->decls= iddecl;				/* insert new id def in decl list */

	idscope->def= def;
	idscope->next= ScopeList->idlist;	/* push idlist of current scope */
	ScopeList->idlist= idscope;			/* insert new id def in current scope */

#ifdef DEBUG
	printf("insert: ");
	print_definition(def);
#endif /* DEBUG */
}

static def_ptr
copy_definition(def_ptr src)
{
	def_ptr	dest;

	if (src == undeclared || src == null_def)
		return src;

	if ((dest= new_definition()) != undeclared) {
		if (src->name)
			dest->name= strdup(src->name);

		dest->sn= src->sn;
		dest->dva= src->dva;
		dest->kind= src->kind;
		dest->type= src->type;
		dest->access= src->access;
		dest->set_or_use= src->set_or_use;

		if (src->bound) {
			int		i;
			dest->dim= 0;			/* will be set by add_def_bound() */
			for (i= 0; i < src->dim; i++) {
				int		low= src->bound[i].low;
				int		size= src->bound[i].size;

				add_def_bound(dest, low, low+size-1);
			}
		} else
			dest->dim= src->dim;
		dest->nchildren= src->nchildren;
	}
#ifdef DEBUG
	printf("copy:   ");
	print_definition(dest);
#endif /* DEBUG */
	return dest;
}

static def_ptr
copy_def_tree(def_ptr src)
{
	def_ptr		dest;

	if (src == undeclared)
		return undeclared;

	dest= copy_definition(src);
	dest->child= copy_def_tree(src->child);
	dest->sibling= copy_def_tree(src->sibling);
	return dest;
}

void
add_def_child(def_ptr parent, def_ptr child)
{
	def_ptr	cdef;

	if (child == undeclared || parent == undeclared)
		return;

	if (parent->child != undeclared) {
		cdef= parent->child;
		while (cdef->sibling != undeclared)
			cdef= cdef->sibling;
		cdef->sibling= child;
	} else
		parent->child= child;

	parent->nchildren++;
	child->set_or_use= 0;

	/* To postpone deletion of child definitions until the moment
	** the parent definition is deleted: */
	child->delete= false;
}

void
find_definition(def_ptr *def, string name)
{
	IDNODE	*idnode;
	def_ptr	src;

	if ((idnode= search(IdTree, name)) == 0 || idnode->decls == 0) {
		i_report(i_ENOTFOUND, name);
		*def= undeclared;
		return;
	}
	src= idnode->decls->def;
	*def= copy_definition(src);
	(*def)->child= copy_def_tree(src->child);

#ifdef DEBUG
	printf("find:   ");
	print_definition(*def);
#endif /* DEBUG */
}

void
find_field(def_ptr *def, string name, def_ptr recdef)
{
	def_ptr	cdef;

	if (recdef == undeclared) {
		*def= undeclared;
		return;
	}

	for (cdef= recdef->child; cdef != undeclared; cdef= cdef->sibling)
		if (cdef->name && !strcmp(cdef->name, name)) {
			*def= copy_definition(cdef);
			use_definition(recdef);
			return;
		}

	i_report(i_ENOFIELD, name, recdef->name);
	*def= undeclared;
}

static void
fill_definition(def_ptr def, int sn, int dpl)
{
	if (def == undeclared || def == null_def)
		return;

	def->sn= sn;
	if (def->nchildren > 0 || def->dva == no_dva)
		def->dva= dpl;
	else
		def->dva += dpl;

	if (def->dim > 0)
		def->kind= (def->type == no_type) ? rec_kind : avar_kind;
	
	if (def->kind == const_kind || def->kind == proc_kind
	|| def->kind == tdecl_kind)
		def->access= readonly_access;
	else
		def->access= direct_access;

	fill_definition(def->child, sn, dpl);
	fill_definition(def->sibling, sn, dpl);
}

void
of_type(def_ptr *def, def_ptr typdef, int dpl)
{
	integer	sn;

	if ((*def= copy_def_tree(typdef)) == undeclared)
		return;

	current_scope(&sn);
	fill_definition(*def, sn, dpl);

	use_definition(typdef);
}

void
get_def_dim(integer *dest, def_ptr def)
{
	*dest= def->dim;
}

void
get_def_type(e_type *dest, def_ptr def)
{
	*dest= def->type;
}

void
get_def_kind(e_kind *dest, def_ptr def)
{
	*dest= def->kind;
}

void
get_def_access(e_access *dest, def_ptr def)
{
	*dest= def->access;
}

void
put_def_dim(integer d, def_ptr def)
{
	if (def != undeclared && def != null_def)
		def->dim= d;
}

void
put_def_type(e_type t, def_ptr def)
{
	if (def != undeclared && def != null_def)
		def->type= t;
}

void
put_def_kind(e_kind k, def_ptr def)
{
	if (def != undeclared && def != null_def) {
		def->kind= k;
		if (k == const_kind || k == proc_kind || k == tdecl_kind)
			put_def_access(readonly_access, def);
		else
			put_def_access(unknown_access, def);
	}
}

void
put_def_access(e_access a, def_ptr def)
{
	if (def != undeclared && def != null_def) {
		if (a == unknown_access)
			def->access &= ~readonly_access;
		else if (a == readonly_access)
			def->access |= a;
		else if ((def->access&readonly_access) == readonly_access)
			def->access= (a|readonly_access);
		else
			def->access= a;
	}
}

void
list_definition(def_ptr def)
{
	i_report(i_TPARSTST, i_def_ptr_text(def));

	for (def= def->child; def != undeclared; def= def->sibling)
		i_report(i_TPARSTST, i_def_ptr_text(def));
}

void
use_definition(def_ptr def)
{
	if (def->set_or_use)
		*(def->set_or_use)= true;
}

static void
printtree(IDNODE *root)
{
	if (root) {
		printtree(root->left);
		if (root->decls)
			list_definition(root->decls->def);
		printtree(root->right);
	}
}

void
list_scope(void)
{
	char	*msg= "All definitions available within scope level %d";

	i_report(i_TPARSTST, msg, ScopeList->depth);
	printtree(IdTree);
}
