/*******************************************************************************
* Implementation of code generation actions for the SLADE VIM code interpreter,
* used in a SLADE parser specification
*
* JHG sep97 Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#include "vimcode.h"

#define MAX_LENTAB	250
#define MAX_STRINGS	250	/* maximum number of strings */

static FILE			*VimFp;
static integer		IC, CurLabel, NrLabels, LenTabSize, StringTabSize;
static integer		LenTab[MAX_LENTAB];
static char			*StringTab[MAX_STRINGS];

char *
i_operation_text(operation op)
{
	switch (op) {
	case abs_:	return "abs_";
	case add:	return "add";
	case and:	return "and";
	case dvi:	return "dvi";
	case eq:	return "eq";
	case ge:	return "ge";
	case gt:	return "gt";
	case jiff:	return "jiff";
	case jift:	return "jift";
	case jump:	return "jump";
	case le:	return "le";
	case lt:	return "lt";
	case mdl:	return "mdl";
	case mul:	return "mul";
	case ne:	return "ne";
	case neg:	return "neg";
	case not:	return "not";
	case or:	return "or";
	case pop:	return "pop";
	case sub:	return "sub";
	case swap:	return "swap";
	default:	;
	}
	return "noop";
}

#ifdef Debugger
void
i_reset_vimcode(void)
{
	int		i;

	if (VimFp) {
		fclose(VimFp);
		VimFp= 0;
	}
	IC= 0;
	NrLabels= 0;
	CurLabel= 0;

	for (i= 0; i < LenTabSize; i++)
		LenTab[i]= 0;
	LenTabSize= 0;

	for (i= 0; i < StringTabSize; i++)
		if (StringTab[i]) {
			free(StringTab[i]);
			StringTab[i]= 0;
		}
	StringTabSize= 0;
}
#endif /* Debugger */

void int_repr_bool(integer *dest, boolean src) { *dest= src; }
void int_repr_char(integer *dest, char src) { *dest= src; }

void
int_repr_string(integer *dest, string src)
{
	if (StringTabSize >= MAX_STRINGS) {
		i_report(i_ESTRTAB, MAX_STRINGS);
		*dest= 0;
		return;
	}
	if ((StringTab[StringTabSize]= strdup(src)) == 0) {
		i_report(i_ENOMEM, __LINE__, __FILE__);
		*dest= 0;
		return;
	}
#ifdef Debugger
	i_add_list_item(IdStringTab, "%-3d %s", StringTabSize, src);
#endif /* Debugger */
	*dest= ++StringTabSize;
}

static void
InstrLine(int nargs, ...)
{
	static int	linenr= 0;
	va_list		args;
	char		*opcode;
	int			arg1, arg2, arg3;

	if (VimFp == 0) {
		if (linenr == 0) {
			i_report(i_EFINVIM);
			linenr= -1;
		}
		return;
	}

	va_start(args, nargs);
	opcode= va_arg(args, char *);
	if (CurLabel != 0)
		fprintf(VimFp, "   %3d:", CurLabel);
	else
		fputs("       ", VimFp);
	IC++;

	switch (nargs) {
	case 3:
		arg1= va_arg(args, int);
		arg2= va_arg(args, int);
		arg3= va_arg(args, int);
		fprintf(VimFp, "%9s %3d %3d %3d", opcode, arg1, arg2, arg3);

		if (linenr != i_linenr) 
			fprintf(VimFp, "%10s %d\n", "# line", i_linenr);
		else
			fputc('\n', VimFp);
#ifdef Debugger
		if (CurLabel != 0)
			i_add_list_item(IdInstr, "%3d:%9s %3d %3d %3d"
			, CurLabel, opcode, arg1, arg2, arg3);
		else
			i_add_list_item(IdInstr, "%13s %3d %3d %3d"
			, opcode, arg1, arg2, arg3);
#endif /* Debugger */
	break;

	case 2:
		arg1= va_arg(args, int);
		arg2= va_arg(args, int);
		fprintf(VimFp, "%9s %3d %3d", opcode, arg1, arg2);

		if (linenr != i_linenr) 
			fprintf(VimFp, "%14s %d\n", "# line", i_linenr);
		else
			fputc('\n', VimFp);
#ifdef Debugger
		if (CurLabel != 0)
			i_add_list_item(IdInstr, "%3d:%9s %3d %3d"
			, CurLabel, opcode, arg1, arg2);
		else
			i_add_list_item(IdInstr, "%13s %3d %3d", opcode, arg1, arg2);
#endif /* Debugger */
	break;

	case 1:
		arg1= va_arg(args, int);
		fprintf(VimFp, "%9s %3d", opcode, arg1);

		if (linenr != i_linenr) 
			fprintf(VimFp, "%18s %d\n", "# line", i_linenr);
		else
			fputc('\n', VimFp);
#ifdef Debugger
		if (CurLabel != 0)
			i_add_list_item(IdInstr, "%3d:%9s %3d", CurLabel, opcode, arg1);
		else
			i_add_list_item(IdInstr, "%13s %3d", opcode, arg1);
#endif /* Debugger */
	break;

	default:									/* no arguments */
		fprintf(VimFp, "%9s", opcode);

		if (linenr != i_linenr) 
			fprintf(VimFp, "%22s %d\n", "# line", i_linenr);
		else
			fputc('\n', VimFp);
#ifdef Debugger
		if (CurLabel != 0)
			i_add_list_item(IdInstr, "%3d:%9s", CurLabel, opcode);
		else
			i_add_list_item(IdInstr, "%13s", opcode);
#endif /* Debugger */
	}
	va_end(args);

	if (CurLabel != 0)
		CurLabel= 0;

	linenr= i_linenr;
}

void
emit_label(integer labnr)
{
	if (CurLabel != 0)
		InstrLine(0, "noop");
	CurLabel= labnr;
}

void
enter_length(integer index, integer len)
{
#ifdef Debugger
	i_add_list_item(IdLenTab, "%-3d %3d", index, len);
#endif /* Debugger */
	LenTab[index - 1]= len;
}

void
get_label(integer *labelnr)
{
	*labelnr= ++NrLabels;
}

void
get_index(integer *index)
{
	if (LenTabSize < MAX_LENTAB)
		*index= ++LenTabSize;
	else
		i_report(i_ELENTAB, MAX_LENTAB);
}

void
assign_op(operation *dest, operation source)
{
	*dest= source;
}

void
clear_op(operation *dest)
{
	*dest= noop;
}

void
emit_op(operation op)
{
	switch (op) {
	case abs_:	InstrLine(0, "abs_"); break;
	case add:	InstrLine(0, "add"); break;
	case and:	InstrLine(0, "and"); break;
	case dvi:	InstrLine(0, "dvi"); break;
	case eq:	InstrLine(0, "eq"); break;
	case ge:	InstrLine(0, "ge"); break;
	case gt:	InstrLine(0, "gt"); break;
	case le:	InstrLine(0, "le"); break;
	case lt:	InstrLine(0, "lt"); break;
	case mdl:	InstrLine(0, "mdl"); break;
	case mul:	InstrLine(0, "mul"); break;
	case ne:	InstrLine(0, "ne"); break;
	case neg:	InstrLine(0, "neg"); break;
	case noop:	InstrLine(0, "noop"); break;
	case not:	InstrLine(0, "not"); break;
	case or:	InstrLine(0, "or"); break;
	case pop:	InstrLine(0, "pop"); break;
	case sub:	InstrLine(0, "sub"); break;
	case swap:	InstrLine(0, "swap"); break;
	default:	;
		i_report(i_EOPARGS, i_operation_text(op));
	}
}

static int
get_def_size(def_ptr def)
{
	int		size;

	if (def->child != undeclared) {
		def_ptr	cdef;

		size= 0;
		for (cdef= def->child; cdef != undeclared; cdef= cdef->sibling)
			size += get_def_size(cdef);
	} else if (def->dim > 0) {
		if (def->bound) {				/* record or static array */
			int		i;

			size= 1;
			for (i= 0; i < def->dim; i++)
				size *= def->bound[i].size;
		} else
			size= 0;					/* dynamic array */
	} else
		size= 1;

	return size;
}

void
update_dpl(integer *newdpl, def_ptr def)
{
	int		size;

	if ((size= get_def_size(def)) == 0)
		*newdpl= def->dva + 2*def->dim + 2;
	else
		*newdpl= def->dva + size;
}

void
emit_opn(operation op, def_ptr def)
{
	if (def->dim > 0) {
		int		size;

		if ((size= get_def_size(def)) != 0)
			InstrLine(1, "ldcon", size);		/* record or static array */
		else
			InstrLine(1, "ldcon", 0);		/* dynamic array */

		switch (op) {
		case eq:
			InstrLine(2, "eqn", def->sn, def->dva);
			break;

		case ne:
			InstrLine(2, "nen", def->sn, def->dva);
			break;

		case pop:
			InstrLine(2, "popn", def->sn, def->dva);
			break;

		default:
			i_report(i_ENOMDOP, i_operation_text(op));
		}
	} else
		emit_op(op);
}

void
emit_descr(int sn, int dpl, int dim)
{
	InstrLine(3, "descr", sn, dpl, dim);
}

void
emit_ldcon(integer arg)
{
	InstrLine(1, "ldcon", arg);
}

void
emit_bound(integer dim, def_ptr def)
{
	int		i= def->dim - dim;

	if (def->bound && i >= 0 && i < def->dim) {
		InstrLine(1, "ldcon", def->bound[i].size);
		InstrLine(1, "ldcon", def->bound[i].low);
	}
}

static void
emit_dimension(def_ptr def)
{
	if (def->bound)			
		InstrLine(1, "ldcon", def->dim);			/* record or static array */
	else
		InstrLine(1, "ldcon", -1);					/* dynamic array */
}

void
emit_xvaraddr(def_ptr def)
{
	emit_dimension(def);
	InstrLine(2, "xvaraddr", def->sn, def->dva);
}

void
emit_swap_pop(def_ptr def, boolean truthval, e_access a)
{
	if (a == unknown_access)
		a= (def->access & ~readonly_access);

	if (truthval
	&& a != address_access
	&& (def->access&indirect_access) == indirect_access) {
		InstrLine(0, "swap");
		InstrLine(0, "pop");
	}
}

void
emit_load(def_ptr def, e_access a, boolean setcheck)
{
	char	*name;

	if (def == undeclared)
		return;

	name= def->name ? def->name : "Expression";

	if (a == unknown_access)
		a= (def->access & ~readonly_access);

	switch (a) {
	case address_access:
		if ((def->access&readonly_access) == readonly_access) 
			i_report(i_ENOADDR, name);
		chk_def_dim(0, def, true);
		chk_def_kind(ptr_kind, def, false);
		use_definition(def);
		break;

	case indirect_access:
		if (def == null_def
		|| (def->kind != ptr_kind
		&& def->kind != rpar_kind
		&& def->kind != avar_kind))
			i_report(i_ENOIND, name);
		break;

	default:
		;
	}

	if (def == null_def) {
		emit_ldcon(def->dva);
		return;
	}

	switch (def->kind) {
	case const_kind:
		emit_ldcon(def->dva);
		use_definition(def);
		break;

	case svar_kind:
	case vpar_kind:
		if (a == address_access)
			InstrLine(2, "varaddr", def->sn, def->dva);
		else
			InstrLine(2, "ldvar", def->sn, def->dva);
		break;

	case rpar_kind:
	case ptr_kind:
		InstrLine(2, "ldvar", def->sn, def->dva);
		if (a == indirect_access)
			InstrLine(0, "ldind");
		break;

	case avar_kind:
		if (def->dim == 0) {
			if (a == direct_access) {
				emit_dimension(def);
				InstrLine(2, "ldxvar", def->sn, def->dva);
			} else if (a == address_access
			&& (def->access&indirect_access) != indirect_access)
				emit_xvaraddr(def);
			else if (a == indirect_access)
				InstrLine(0, "ldind");
		} else {
			int		size;
			if ((size= get_def_size(def)) != 0)
				InstrLine(1, "ldcon", size);		/* record or static array */
			else
				InstrLine(1, "ldcon", 0);		/* dynamic array */

			InstrLine(2, "ldnvar", def->sn, def->dva);
		}
		break;

	case rec_kind:
		if (def->dim == 0) {
			if (a == address_access)
				InstrLine(2, "varaddr", def->sn, def->dva);
			else
				InstrLine(2, "ldvar", def->sn, def->dva);
		} else {
			int		size;
			if ((size= get_def_size(def)) != 0)
				InstrLine(1, "ldcon", size);		/* record or static array */
			else
				InstrLine(1, "ldcon", 0);		/* dynamic array */

			InstrLine(2, "ldnvar", def->sn, def->dva);
		}
		break;

	default:
		chk_def_kind(proc_kind, def, false);
		chk_def_kind(tdecl_kind, def, false);
	}

	if (setcheck && def->set_or_use && !*(def->set_or_use)) {
		i_report(i_WNOTSET, def->name);
		use_definition(def);
	}
}

void
emit_store(def_ptr def, boolean load)
{
	char	*name;

	if (def == undeclared)
		return;

	name= def->name ? def->name : "expression";

	use_definition(def);

	if ((def->access&readonly_access) == readonly_access)
		i_report(i_ENOSTORE, name);

	if ((def->access&address_access) == address_access)
		i_report(i_EADDRESS, name);

	switch (def->kind) {
	case svar_kind:
	case vpar_kind:
		InstrLine(2, "stvar", def->sn, def->dva);
		break;

	case rpar_kind:
	case ptr_kind:
		if ((def->access&indirect_access) == indirect_access)
			InstrLine(0, "stind");
		else
			InstrLine(2, "stvar", def->sn, def->dva);
		break;

	case avar_kind:
		if (def->dim == 0) {
			if ((def->access&direct_access) == direct_access) {
				emit_dimension(def);
				InstrLine(2, "stxvar", def->sn, def->dva);
			} else if ((def->access&indirect_access) == indirect_access)
				InstrLine(0, "stind");
		} else {
			int		size;
			if ((size= get_def_size(def)) != 0)
				InstrLine(1, "ldcon", size);		/* record or static array */
			else
				InstrLine(1, "ldcon", 0);		/* dynamic array */

			InstrLine(2, "stnvar", def->sn, def->dva);
		}
		break;

	case rec_kind:
		if (def->dim == 0)
			InstrLine(2, "stvar", def->sn, def->dva);
		else {
			int		size;
			if ((size= get_def_size(def)) != 0)
				InstrLine(1, "ldcon", size);		/* record or static array */
			else
				InstrLine(1, "ldcon", 0);		/* dynamic array */

			InstrLine(2, "stnvar", def->sn, def->dva);
		}
		break;

	default:
		if ((def->access&readonly_access) != readonly_access)
			i_report(i_ENOSTORE, name);
	}
	if (load) {
		if ((def->access&indirect_access) == indirect_access)
			InstrLine(0, "ldind");
		else
			emit_load(def, unknown_access, true);
	}
}

static void
emit_starg(def_ptr cdef)
{
	if (cdef->sibling != undeclared)
		emit_starg(cdef->sibling);	/* recursion results in reverse order */

	if (cdef->kind == rpar_kind)
		put_def_access(direct_access, cdef);

	emit_store(cdef, false);

	if (cdef->kind == rpar_kind)
		put_def_access(indirect_access, cdef);
}

void
emit_stargs(def_ptr def)
{
	if (def->kind == proc_kind && def->child != undeclared)
		emit_starg(def->child);
}

void
emit_jump(operation op, integer labnr)
{
	switch (op) {
	case jiff:	InstrLine(1, "jiff", labnr); break;
	case jift:	InstrLine(1, "jift", labnr); break;
	case jump:	InstrLine(1, "jump", labnr); break;
	default:
		i_report(i_EILLARG, i_operation_text(op));
	}
}

void
emit_call(def_ptr def)
{
	if (def == undeclared || def == null_def)
		return;

	chk_def_kind(proc_kind, def, true);
	use_definition(def);
	InstrLine(1, "call", def->dva);
}

void
emit_return()
{
	InstrLine(0, "return_");
}

void
emit_read(e_type t)
{
	switch (t) {
	case bool_type:
		InstrLine(0, "rdbool");
		break;
	case char_type:
		InstrLine(0, "rdchar");
		break;
	case string_type:
		InstrLine(0, "rdstring");
		break;
	case int_type:
		InstrLine(0, "rdint");
		break;
	default:
		i_report(i_ERDTYPE, i_e_type_text(t));
	}
}

void
emit_write(e_type t)
{
	switch (t) {
	case bool_type:
		InstrLine(0, "wrbool");
		break;
	case char_type:
		InstrLine(0, "wrchar");
		break;
	case string_type:
		InstrLine(0, "wrstring");
		break;
	case int_type:
		InstrLine(0, "wrint");
		break;
	default:
		i_report(i_EWRTYPE, i_e_type_text(t));
	}
}

void
emit_crseg(integer level, integer lenindex)
{
	InstrLine(2, "crseg", level, lenindex);
}

void
emit_dlseg()
{
	InstrLine(0, "dlseg");
}

void
initialise_vimcode(void)
{
	integer	i;
	string	filename;

	if (VimFp) {
		i_report(i_EINIVIM);
		return;
	}

	sprintf(filename, "%s.vim", i_input_name);
	if ((VimFp= fopen(filename, "wt")) == 0) {
		i_report(i_EFWRITE, filename);
		return;
	}

	i_report(i_TVIM, filename);

	fprintf(VimFp, "# VIM code '%s'\n#\n", filename);
	fputs("# Label   Opcode Arguments\n#\n", VimFp);

	IC= 0;
	LenTabSize= 0;
	StringTabSize= 0;
	NrLabels= 0;
	CurLabel= 0;

	for (i= 0; i < MAX_LENTAB; i++)
		LenTab[i]= 0;
}

void
finalise_vimcode(void)
{
	integer	i;

	if (VimFp == 0) {
		i_report(i_EFINVIM);
		return;
	}

	i_report(i_TNITAB, IC);
	i_report(i_TNLTAB, LenTabSize);
	i_report(i_TNSTAB, StringTabSize);

	InstrLine(0, "halt");
	fprintf(VimFp, "#\n# end of VIM code\n#\n");

	fputs(".\n\n", VimFp);
	if (LenTabSize > 0) {
		fputs("# length table\n", VimFp);
		fputs("# index length\n", VimFp);
		for (i= 0; i < LenTabSize; i++) {
			fprintf(VimFp, "   %3d: %d\n", i+1, LenTab[i]);
			LenTab[i]= 0;
		}
		fprintf(VimFp, "#\n# end of length table\n#\n");
		LenTabSize= 0;
	} else
		fputs("# length table empty\n", VimFp);

	fputs(".\n\n", VimFp);
	if (StringTabSize > 0) {
		fputs("# string table\n", VimFp);
		fputs("# index string\n", VimFp);
		for (i= 0; i < StringTabSize; i++) {
			fprintf(VimFp, "   %3d: \"%s\"\n", i+1, StringTab[i]);
			free(StringTab[i]);
			StringTab[i]= 0;
		}
		fprintf(VimFp, "#\n# end of string table\n#\n");
		StringTabSize= 0;
	} else
		fputs("# string table empty\n", VimFp);

	fclose(VimFp);
	VimFp= 0;
}

void dy_op_bool(boolean *dest, operation op, boolean src)
{
	switch (op) {
	case or:	*dest= (*dest || src);			break;
	case and:	*dest= (*dest && src);			break;
	case gt:	*dest= (*dest > src);			break;
	case ge:	*dest= (*dest >= src);			break;
	case lt:	*dest= (*dest < src);			break;
	case le:	*dest= (*dest <= src);			break;
	case eq:	*dest= (*dest == src);			break;
	case ne:	*dest= (*dest != src);			break;
	case add:	*dest= (*dest || src);			break;
	case sub:	*dest= (*dest != src);			break;
	case mul:	*dest= *dest * src;				break;
	case mdl:
		if (src)
			*dest= *dest % src;
		else
			i_report(i_EMODBY0);
		break;

	case dvi:
		if (src != 0)
			*dest= *dest / src;
		else
			i_report(i_EDIVBY0);
		break;

	case neg:	*dest= -src;					break;
	case not:	*dest= !src;					break;
	default:	*dest= src;
	}
}

void dy_op_char(char *dest, operation op, char src)
{
	switch (op) {
	case or:	*dest= (*dest || src);	break;
	case and:	*dest= (*dest && src);	break;
	case gt:	*dest= (*dest > src);	break;
	case ge:	*dest= (*dest >= src);	break;
	case lt:	*dest= (*dest < src);	break;
	case le:	*dest= (*dest <= src);	break;
	case eq:	*dest= (*dest == src);	break;
	case ne:	*dest= (*dest != src);	break;
	case add:	*dest= (*dest || src);	break;
	case sub:	*dest= (*dest != src);	break;
	case mul:	*dest= *dest * src;	break;
	case mdl:
		if (src)
			*dest= *dest % src;
		else
			i_report(i_EMODBY0);
		break;

	case dvi:
		if (src != 0)
			*dest= *dest / src;
		else
			i_report(i_EDIVBY0);
		break;

	case neg:	*dest= -src;					break;
	case not:	*dest= !src;					break;
	default:
		*dest= src;
	}
}

void dy_op_int(integer *dest, operation op, integer src)
{
	switch (op) {
	case or:	*dest= (*dest || src);			break;
	case and:	*dest= (*dest && src);			break;
	case gt:	*dest= (*dest > src);			break;
	case ge:	*dest= (*dest >= src);			break;
	case lt:	*dest= (*dest < src);			break;
	case le:	*dest= (*dest <= src);			break;
	case eq:	*dest= (*dest == src);			break;
	case ne:	*dest= (*dest != src);			break;
	case add:	*dest= *dest + src;				break;
	case sub:	*dest= *dest - src;				break;
	case mul:	*dest= *dest * src;				break;
	case mdl:
		if (src)
			*dest= *dest % src;
		else
			i_report(i_EMODBY0);
		break;

	case dvi:
		if (src != 0)
			*dest= *dest / src;
		else
			i_report(i_EDIVBY0);
		break;

	case neg:	*dest= -src;					break;
	case abs_:	*dest= (src < 0) ? -src : src;	break;
	case not:	*dest= !src;					break;
	default:	*dest= src;
	}
}

void dy_op_real(real *dest, operation op, real src)
{
	switch (op) {
	case or:	*dest= (*dest != 0.0 || src != 0.0);	break;
	case and:	*dest= (*dest != 0.0 && src != 0.0);	break;
	case gt:	*dest= (*dest > src);					break;
	case ge:	*dest= (*dest >= src);					break;
	case lt:	*dest= (*dest < src);					break;
	case le:	*dest= (*dest <= src);					break;
	case eq:	*dest= (*dest == src);					break;
	case ne:	*dest= (*dest != src);					break;
	case add:	*dest= *dest + src;						break;
	case sub:	*dest= *dest - src;						break;
	case mul:	*dest= *dest * src;						break;
	case dvi:
		if (src != 0.0)
			*dest= *dest / src;
		else
			i_report(i_EDIVBY0);
		break;

	case neg:	*dest= -src;							break;
	case abs_:	*dest= (src < 0.0) ? -src : src;		break;
	case not:	*dest= !src;							break;
	default:	*dest= src;
	}
}

void mon_op_bool(boolean *dest, operation op)
{
	switch (op) {
	case neg:	*dest= -(*dest);		break;
	case not:	*dest= !(*dest);		break;
	default:	*dest= *dest;
	}
}

void mon_op_char(char *dest, operation op)
{
	switch (op) {
	case neg:	*dest= -(*dest);		break;
	case not:	*dest= !(*dest);		break;
	default:	*dest= *dest;
	}
}

void mon_op_int(integer *dest, operation op)
{
	switch (op) {
	case neg:	*dest= -(*dest);						break;
	case abs_:	*dest= (*dest < 0) ? -(*dest) : *dest;	break;
	case not:	*dest= !(*dest);						break;
	default:	*dest= *dest;
	}
}

void mon_op_real(real *dest, operation op)
{
	switch (op) {
	case neg:	*dest= -(*dest);						break;
	case abs_:	*dest= (*dest < 0.0) ? -(*dest) : *dest;break;
	case not:	*dest= !(*dest);						break;
	default:	*dest= *dest;
	}
}
